mod create;
pub(crate) mod load;
pub(crate) mod store;

#[derive(Debug)]
pub enum ModelError {
    DBError(rusqlite::Error),
    LoadError(String),
    EmptyStoreError,
}

impl From<rusqlite::Error> for ModelError {
    fn from(e: rusqlite::Error) -> Self {
        Self::DBError(e)
    }
}

impl From<ModelError> for rusqlite::Error {
    fn from(e: ModelError) -> Self {
        match e {
            ModelError::DBError(e) => e,
            _ => panic!()
        }
    }
}

/*impl Into<rusqlite::Error> for ModelError {
    fn into(self) -> rusqlite::Error {
        match self {
            Self::DBError(e) => e,
            _ => panic!()
        }
    }
}*/

impl std::fmt::Display for ModelError {
    fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        fmt.write_fmt(format_args!("{:?}", self))
    }
}

impl serde::ser::Error for ModelError {
    fn custom<T: std::fmt::Display>(msg: T) -> Self {
        Self::LoadError(format!("{}", msg))
    }
}

impl serde::de::Error for ModelError {
    fn custom<T: std::fmt::Display>(msg: T) -> Self {
        Self::LoadError(format!("{}", msg))
    }
}

impl std::error::Error for ModelError {}

/// A database entity, aka a struct representing a row in a table
pub trait Entity: for<'de> serde::Deserialize<'de> + serde::Serialize {
    type Column;
    fn table_name() -> &'static str;
    fn column_count() -> usize
    where
        Self: Sized;
    fn index(c: Self::Column) -> usize
    where
        Self: Sized;
    fn name(c: Self::Column) -> &'static str
    where
        Self: Sized;
    fn values(&self) -> Vec<&dyn rusqlite::ToSql>;
}

pub trait EntityColumns {
    type Entity;
}

/// How we describe an entire schema
#[derive(Debug)]
pub struct SchemaModel {
    drop: Vec<String>,
    create: Vec<String>,
}

impl SchemaModel {
    pub fn new() -> Self {
        Self {
            drop: Vec::new(),
            create: Vec::new(),
        }
    }

    pub fn add<E: Entity>(mut self) -> Self {
        let (drop, create) = create::sql_for::<E>();
        self.drop.push(drop);
        self.create.push(create);
        self
    }

    pub fn drop(&self) -> &Vec<String> {
        &self.drop
    }
    pub fn create(&self) -> &Vec<String> {
        &self.create
    }
}

impl Default for SchemaModel {
    fn default() -> Self {
        Self::new()
    }
}
